(ns kube-deploy.core
  (:require [clojure.java.io :as io]
            [clojure.java.shell :as sh]
            [cli-matic.core :refer [run-cmd]]
            [yaml.core :as yaml]
            [cheshire.core :as json]
            [kube-deploy.from-yaml :as lky]
            [kube-deploy.data :as data]
            [kube-deploy.helpers :as lkh]
            [lambdakube.core :as lk]
            [kube-deploy.blocks :as block]
            [lambdakube.util :as lku]))

(def conf (atom {}))
(def services (atom {}))
(reset! services (clojure.edn/read-string (slurp "config.edn")))
(def deployment-config {:num-be-slaves 2 :num-fe-replicas 2})

(def resources-mini
  {:requests
   {:cpu "100m" :memory "64Mi"}
   :limits
   {:cpu "200m" :memory "256Mi"}})

(def resources-small
  {:requests
   {:cpu "100m" :memory "100Mi"}
   :limits
   {:cpu "200m" :memory "400Mi"}})

(def resources-medium
  {:requests
   {:cpu "100m" :memory "500Mi"}
   :limits
   {:cpu "200m" :memory "1Gi"}})

; will not be needed soon

(defn kube-get-namespaces [kube-config]
  (into #{} (->> (lkh/kube-namespaces kube-config)
                 (map :metadata)
                 (map :name))))

(defn kube-does-namespace-exists
  ([namespace]
   (kube-does-namespace-exists namespace nil))
  ([namespace kube-config]
   (some #(= (name namespace) %) (kube-get-namespaces kube-config))))




; will not be needed soon


(defn kube-apply
  ([content file]
   (kube-apply content file nil nil))
  ([content file kube-namespace]
   (kube-apply content kube-namespace nil))
  ([content file kube-namespace kube-config]
   (let [namespace-param (if (nil? kube-namespace) nil (str "--namespace=" kube-namespace))
         config-param (if (nil? kube-config) nil (str "--kubeconfig=" kube-config))]
     (when-not (and (.exists file)
                    (= (slurp file) content))
       (spit file content)
       (let [res (apply sh/sh (remove nil? ["kubectl" "apply" namespace-param config-param "-f" (str file)]))]
         (when-not (= (:exit res) 0)
           (.delete file)
           (throw (Exception. (:err res)))))))))

(defn redis [kube-injector]
  (-> kube-injector
      (lk/rule :redis-master-rule []
               (fn []
                 (-> (lk/pod :redis-master {:app :footprint
                                            :env :production
                                            :role :master
                                            :tier :redis})
                     (lk/add-container :redis (get-in @conf [:images :redis])
                                       (-> {}
                                           (lkh/add-to-spec :resources resources-small)
                                           (lkh/add-readiness-probe-cmd ["redis-cli" "--version"])
                                           (lkh/add-liveness-probe-cmd ["redis-cli" "--version"])))
                     (lk/deployment 1)
                     (lk/expose-cluster-ip :redis-master
                                           (lk/port :redis :redis 6379 6379)))))

      #_(lk/rule :redis-slave-rule [:redis-master-rule :num-be-slaves]
                 (fn [backend-master num-be-slaves]
                   (-> (lk/pod :redis-slave {:app :redis
                                             :role :slave
                                             :tier :backend})
                       (lk/add-container :redis (get-in @conf [:images :redis])
                                         (-> {}
                                             (lkh/add-to-spec :resources resources-small)
                                             (lkh/add-to-spec
                                              :args ["--slaveof"
                                                     (:hostname backend-master)
                                                     (-> backend-master :ports :redis str)])
                                             (lkh/add-liveness-probe-cmd ["redis-cli" "--version"])))
                       (lk/deployment num-be-slaves)
                       (lk/expose-cluster-ip :redis-slave
                                             (lk/port :redis :redis 6379 6379)))))))

(defn mosquitto-mqtt [kube-injector]
  (-> kube-injector
      (lk/rule :mosquitto-master-rule []
               (fn []
                 (-> (lk/pod
                      :mosquitto-master
                      {:app :footprint :role :master :env :production :tier :database})
                     (lk/add-container :mosquitto "eclipse-mosquitto"
                                       (-> {}
                                           (lkh/add-to-spec :resources resources-mini)
                                           (lkh/add-readiness-probe-cmd ["pg_isready"])
                                           (lkh/add-liveness-probe-cmd ["pg_isready"])
                                           (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])))
                     (lk/deployment 1)
                     (lk/expose-cluster-ip :mosquitto-master
                                           (lk/port :mosquitto :mosquitto 1883 1883)))))))

(defn database [kube-injector]
  (-> kube-injector
      (lk/rule :postgres-master-rule []
               (fn []
                 (-> (lk/pod
                      :postgres-master
                      {:app :footprint :role :master :env :production :tier :database})
                     (lk/add-container :postgres (get-in @conf [:images :postgres])
                                       (-> {}
                                           (lkh/add-to-spec :resources resources-small)
                                           (lkh/add-readiness-probe-cmd ["pg_isready"])
                                           (lkh/add-liveness-probe-cmd ["pg_isready"])
                                           (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])))
                     (lk/stateful-set 1)

                     (lk/add-volume-claim-template :postgres-volume
                                                   {:accessModes ["ReadWriteOnce"]
                                                    :resources {:requests {:storage "1Gi"}}}
                                                   {:bar "/var/lib/postgresql/data"})
                     (lk/expose-cluster-ip :postgres-master
                                           (lk/port :postgres :postgres 5432 5432)))))))

(defn rabbitmq [kube-injector]
  (-> kube-injector
      (lk/rule :rabbit-master-rule []
               (fn []
                 (-> (lk/pod
                      :rabbit-master
                      {:app :footprint :role :master :env :prodction :tier :rabbit})
                     (lk/add-container
                      :rabbit
                      (get-in @conf [:images :rabbitmq])
                      (-> {}
                          (lkh/add-to-spec :resources resources-medium)
                          (lkh/add-liveness-probe-cmd ["rabbitmq-diagnostics" "ping" "-q"])))

                     (lk/deployment 1)
                     (lk/expose-cluster-ip
                      :rabbit-master (lk/port :rabbit :rabbit 5672 5672)))))))

(defn hackspace [kube-injector]
  (-> kube-injector
      (lk/rule :hackspace-backend-master [:postgres-master-rule :redis-master-rule]
               (fn [postgres redis]
                 (-> (lk/pod :hackspace-master {:app :hackspace
                                                :role :master
                                                :tier :web}
                             {:imagePullSecrets [{:name "docker-registry"}]})

                     ; Add containers to the pod, one for frontend assets and one for backend
                     (lk/add-container
                      :hackspace-backend
                      (get-in @conf [:images :django])
                      (-> {}
                          (lkh/add-to-spec :resources resources-medium)
                          ;(lkh/add-to-spec :command ["python" "/app/wsgi_bjoern.py"])
                          ;(lkh/add-to-spec :command ["/gunicorn.sh"])
                          (lkh/add-to-spec :command ["/usr/local/bin/gunicorn" "config.wsgi" "--workers" "2" "-k" "gevent" "--worker-connections" "100" "--max-requests" "300" "--keep-alive" "1000" "-b" "0.0.0.0:5000" "--chdir=/app"])
                          (lkh/add-to-spec :ports [{:containerPort 5000 :name "gunicorn-port"}])
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])
                          #_(lkh/add-readiness-probe-cmd ["wget" "127.0.0.1:5000"])
                          #_(lkh/add-liveness-probe-cmd ["wget" "127.0.0.1:5000"])))

                     #_(lk/add-container
                        :hackspace-frontend
                        (get-in @conf [:images :nginx])
                        (-> {}
                            (lkh/add-to-spec :resources resources-mini)
                            (lkh/add-to-spec :ports [{:containerPort 80 :name "nginx-port"}])
                            (lkh/add-readiness-probe-cmd ["wget" "127.0.0.1/static/images/favicon.ico"])
                            (lkh/add-liveness-probe-cmd ["wget" "127.0.0.1/static/images/favicon.ico"])))
                     (lk/add-volume "static-files" {:emptyDir {}}
                                    {:hackspace-backend "/app/staticfiles"})

                       ; order does matter here, deployment after initcontainers
                     ;(lku/wait-for-service-port postgres :postgres)
                     ;(lku/wait-for-service-port redis :redis)

                       ; wrap in a deployment
                     (lk/deployment 2)
                     (lk/expose-cluster-ip :hackspace-backend
                                           (lk/port :hackspace-master :gunicorn-port 5000 5000))
                     (lk/expose-cluster-ip :hackspace-frontend
                                           (lk/port :hackspace-master :nginx-port 80 80)))))
      (lk/rule :hackspace-queue-master []
               (fn []
                 (-> (lk/pod :hackspace-queue
                             {:app :hackspace
                              :env :stage
                              :role :master
                              :tier :queue}
                             {:imagePullSecrets [{:name "docker-registry"}]})
                     (lk/add-container
                      :hackspace
                      (get-in @conf [:images :django])
                      (-> {}
                          (lkh/add-to-spec :resources resources-small)
                          (lkh/add-to-spec :command ["/app/manage.py" "run_huey"])
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])))
                     (lk/deployment 2))))))

(defn field-conj [m k v]
  (if (contains? m k)
    (update m k conj v)
    ;; else
    (assoc m k [v])))

(defn- mount-func [name mounts]
  (apply comp (for [[cont path-or-opts] mounts]
                #(lk/update-container % cont field-conj :volumeMounts
                                      (if (map? path-or-opts)
                                        (into path-or-opts {:name name})
                                        {:name name
                                         :mountPath-Or-Opts path-or-opts})))))

(defn add-volume [pod name spec mounts]
  (-> pod
      (update :spec field-conj :volumes (-> {:name name}
                                            (merge spec)))
      ((mount-func name mounts))))

(defn footprint [kube-injector]
  (-> kube-injector
      (lk/rule :footprint-backend-master [:postgres-master-rule :redis-master-rule :rabbit-master-rule]
               (fn [postgres redis rabbit]
                 (-> (lk/pod :footprint-app {:app :footprint
                                             :env :production
                                             :tier :web}
                             {:imagePullSecrets [{:name "docker-registry"}]})

                     ; Add containers to the pod, one for frontend assets and one for backend
                     (lk/add-container
                      :footprint-backend
                      (get-in @conf [:images :footprint])
                      (-> {}
                          (lkh/add-to-spec :resources resources-medium)
                          ;(lkh/add-to-spec :command ["/gunicorn.sh"])
                          (lkh/add-to-spec :command ["/usr/bin/gunicorn" "config.wsgi"  "-b" "0.0.0.0:5000" "--chdir=/app" "--log-file=/tmp/gunicorn.log"])

                          (lkh/add-to-spec :ports [{:containerPort 5000 :name "gunicorn-port"}])
                          (lkh/add-to-spec :imagePullPolicy "Always")
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])
                          #_(lkh/add-readiness-probe-get "http://127.0.0.1" 5000)
                          (lkh/add-liveness-probe-get "http://127.0.0.1" 5000)))

                     (lk/add-container
                      :footprint-frontend
                      (get-in @conf [:images :nginx])
                      (-> {}
                          (lkh/add-to-spec :resources resources-mini)
                          (lkh/add-to-spec :imagePullPolicy "Always")
                          (lkh/add-to-spec :ports [{:containerPort 80 :name "nginx-port"}])
                          #_(lkh/add-readiness-probe-get "127.0.0.1/static/images/favicon.ico" 80)
                          (lkh/add-liveness-probe-get "http://127.0.0.1/static/images/favicon.ico" 80)))

                     (add-volume "static-files" {:emptyDir {}}
                                 {:footprint-backend {:mountPath "/app/staticfiles"}
                                  :footprint-frontend
                                  {:mountPath "/usr/share/nginx/html/static/"
                                   :readOnly true}})

                     (lk/add-init-container
                      :footprint-init-migrations
                      (get-in @conf [:images :footprint])
                      (-> {}
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])
                          (lkh/add-to-spec :command ["./manage.py" "migrate"])))

                     (lk/add-init-container
                      :footprint-init-static-resources
                      (get-in @conf [:images :footprint])
                      (-> {}
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])
                          (lkh/add-to-spec :command ["./manage.py" "collectstatic"])))

                     ; order does matter here, deployment after initcontainers
                     ;(lku/wait-for-service-port postgres :postgres)
                     ;(lku/wait-for-service-port redis :redis)
                     ;(lku/wait-for-service-port rabbit :rabbit)

                     ; wrap in a deployment
                     (lk/deployment 2)
                     (lk/expose-cluster-ip :footprint-backend
                                           (lk/port :footprint-app :gunicorn-port 5000 5000))
                     (lk/expose-cluster-ip :footprint-frontend
                                           (lk/port :footprint-app :nginx-port 80 80)))))
      (lk/rule :footprint-aqscheduler-master []
               (fn []
                 (-> (lk/pod :footprint-scheduler
                             {:app :footprint
                              :env :production
                              :tier :scheduler}
                             {:imagePullSecrets [{:name "docker-registry"}]})

                     (lk/add-container
                      :scheduler
                      (get-in @conf [:images :scheduler])
                      (-> {}
                          (lkh/add-to-spec :imagePullPolicy "Always")
                          (lkh/add-to-spec :resources resources-mini)
                          (lkh/add-to-spec :command ["/usr/local/bin/dramatiq_apscheduler" "/app/tasks.yml" "--debug"])
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])))
                     (add-volume "scheduler-config" {:configMap
                                                     {:name "scheduler-tasks"}}
                                 {:scheduler
                                  {:name "scheduler-config"
                                   :mountPath "/app/tasks.yml"
                                   :subPath "tasks.yml"}})
                     (lk/deployment 1))))

      (lk/rule :footprint-queue-master []
               (fn []
                 (-> (lk/pod :footprint-queue
                             {:app :footprint
                              :env :production
                              :tier :queue}
                             {:imagePullSecrets [{:name "docker-registry"}]})

                     (lk/add-container
                      :footprint
                      (get-in @conf [:images :footprint])
                      (-> {}
                          (lkh/add-to-spec :imagePullPolicy "Always")
                          (lkh/add-to-spec :resources resources-medium)
                          (lkh/add-to-spec :command ["/app/manage.py" "rundramatiq" "-p" "1" "-t" "1"])
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])))

                     (lk/deployment 2))))))

(defn make-secrets [kube-injector name data]
  (-> kube-injector
      (lk/rule :secrets []
               (fn []
                 (-> (lkh/secrets (keyword name) data))))))

(defn make-registry-secrets [kube-injector name data]
  (if (nil? (:.dockerconfigjson data))
    kube-injector
    (-> kube-injector
        (lk/rule
         :secrets-registry []
         (fn []
           (lkh/secrets-registry name {} (lkh/encode-secret-map data)))))))

(defn make-namespace [kube-injector name]
  (lk/rule kube-injector :namespace []
           (fn []
             (lkh/kube-namespace (keyword name)))))

(defn make-environment-secrets [kube-injector name data]
  (-> kube-injector
      (lk/rule :secrets-environment []
               (fn []
                 (lkh/secrets name {}
                              (lkh/encode-secret-map data))))))

(defn add-config [kube-injector name config]
  (lk/rule kube-injector :config-map []
           (fn []
             (lk/config-map name config))))

#_(defn make-ingress-hackspace [kube-injector]
  (lk/rule kube-injector :lkube-ingress []
           (fn []
             (-> (lkh/ingress "lkube-ingress" {:app :nginx-ingress :env :production :tier :ingress})
                 (lkh/add-spec :tls [{:hosts [(:ingress-host @conf)] :secretName "letsencrypt-production"}])
                 (lkh/add-spec :rules [{:host (:ingress-host @conf)
                                        :http {:paths
                                               [{:path "/"
                                                 :backend
                                                 {:serviceName "hackspace-backend"
                                                  :servicePort 5000}}
                                                {:path "/static/"
                                                 :backend
                                                 {:serviceName "hackspace-frontend"
                                                  :servicePort 80}}]}}])
                 (lk/add-annotation :certmanager.k8s.io/cluster-issuer "letsencrypt-production")
                 (lk/add-annotation :kubernetes.io/ingress.class "nginx")))))

#_(defn make-ingress [kube-injector]
  (lk/rule kube-injector :lkube-ingress []
           (fn []
             (-> (lkh/ingress "lkube-ingress" {:app :nginx-ingress :env :production :tier :ingress})
                 (lkh/add-spec :tls [{:hosts [:lkube-ingress] :secretName "lkube-tls"}])
                 (lkh/add-spec :rules [{:host (:ingress-host @conf)
                                        :http {:paths
                                               [{:path "/"
                                                 :backend
                                                 {:serviceName "footprint-backend"
                                                  :servicePort 5000}}
                                                {:path "/static/"
                                                 :backend
                                                 {:serviceName "footprint-frontend"
                                                  :servicePort 80}}]}}])
                 (lk/add-annotation :kubernetes.io/tls-acme "true")
                 (lk/add-annotation :kubernetes.io/ingress.class "nginx")))))

#_(defn footprint-deployment [namespace kube-config]
    (println "deploying footprint")
    (reset! conf (clojure.edn/read-string (slurp (str namespace "-config.edn"))))
  ; run deployment in default namespace to create a namespace called footprint
    (-> (lk/injector)
        (make-namespace namespace)
        lk/standard-descs
        (lk/get-deployable deployment-config)
        (lkh/to-yaml-store (str namespace "-store.yaml"))
        (lk/kube-apply (io/file (str namespace "-deploy.yaml")) nil kube-config))

  ; run deployment under footprint namespace
    (-> (lk/injector)
        (make-environment-secrets :environment (:vars @conf))
        #_(make-registry-secrets :docker-registry {:.dockerconfigjson (:registry @conf)})
        make-ingress
        rabbitmq
        database
        redis
        (add-config "scheduler-tasks" (get-in @conf [:config]))
        footprint
        lk/standard-descs
        (lk/get-deployable deployment-config)
        (lkh/to-yaml-store (str namespace "-store.yaml"))
        (lk/kube-apply (io/file (str namespace "-deploy.yaml")) "footprint" kube-config)))

(defn create-namespace-if-not-exists [{:keys [config namespace blocks]}]
  (when-not (kube-does-namespace-exists (keyword namespace) config)
    (println (str "Namespace " namespace " not found creating"))
    (-> (lk/injector)
        (make-namespace namespace)
        (lk/get-deployable deployment-config)
        (lkh/to-yaml-store (str namespace "-namespace.yaml"))
        (lk/kube-apply (io/file (str namespace "-deploy.yaml")) nil config))))


; recursive function that processes through each project block
(defn apply-blocks [kube-injector func settings]
  (if (some? (first func))
    (do
        (println "Generating block" (first func))
        (recur
         (case (first func)
           :ldap (block/ldap kube-injector)
           :redis (block/redis kube-injector)
           :rabbitmq (block/rabbitmq kube-injector)
           :postgres (block/postgres kube-injector)
           :mosquitto (block/mosquitto-mqtt kube-injector)
           :django-huey (block/django-huey kube-injector (get-in settings [(keyword (first func))]))
           :django-gunicorn (block/gunicorn kube-injector (get-in settings [(keyword (first func))]))
           kube-injector)
         (rest func)
         settings))
         kube-injector))


; build functionality, pass config to apply blocks to generate the map then dump yaml
(defn build [{:keys [config namespace blocks options] :as service}]
  (println "Deploying to" (or config "default") "namespace" namespace)
  (println options)
  (if (.exists (io/file (str namespace "-config.edn")))
    (do (reset! conf (clojure.edn/read-string (slurp (str namespace "-config.edn"))))
        (create-namespace-if-not-exists service)
        (as-> (lk/injector)
              injector
          (if (= "t" (:secrets options))
            (make-environment-secrets injector :environment (:vars @conf))
            injector)
                                        ;(make-registry-secrets injector :docker-registry {:.dockerconfigjson (:registry @conf)})
          (apply-blocks injector blocks @conf)
          (lk/standard-descs injector)
          (lk/get-deployable injector deployment-config)
          (lkh/to-yaml-store injector (str namespace "-store.yaml"))
          (if (= "t" (:deploy options))
            (lk/kube-apply injector (io/file (str namespace "-deploy.yaml")) namespace config))))
    (println "No config found "  (str namespace "-config.yaml"))))

; actual deploy config
(defn deploy [{:keys [namespace] :as ns-config}]
  (println ns-config)
  (let [config (get-in data/services [(keyword namespace) :config])
        blocks (get-in data/services [(keyword namespace) :blocks])]
    (cond
      ;(= namespace "footprint-production") (footprint-deployment "footprint-production" "/home/oly/.kube/ake.yml")
      ;(= namespace "footprint-stage") (footprint-deployment "footprint" "/home/oly/.kube/ake.yml")
      (= (nil? blocks)) (build {:namespace namespace :config config :blocks blocks :options ns-config})
      :else (println (str "Namespace " namespace " not found.")))))

(def CONFIGURATION
  {:app         {:command     "Kubernetes deploy"
                 :description "Generate yaml and deploy out to cluster."
                 :version     "0.0.1"}
;   :global-opts []
   :commands    [{:command     "deploy"
                  :description "Deploy app out to cluster"
                  :opts        [{:short "n" :option "namespace" :as "Namespace to deploy" :type :string :spec string?}
                                {:short "d" :option "deploy" :as "Deploy with kubectl" :type :string :default nil}
                                {:short "s" :option "secrets" :as "Include secrets" :type :string :default nil}]
                  :runs        deploy}
                 {:command     "secrets"
                  :description "Display secrests"
                  :opts        [{:option "service" :as "Service to deploy" :type :string :spec string?}]
                  :runs        lky/fetch-secrets}]})

(defn -main [& args]
  (if args
    (run-cmd args CONFIGURATION)))


